/* 
 * Windmesser - SW to caclulate  Windgust, Wind speed and direction from 
 * Davis Vantage Pro 2 Anemometer
 * 
 * Copyright (c) 2021 W. Artmann.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <TimerOne.h>
#include<Wire.h>
#include<math.h>

#include<ADS1115_WE.h> 
#include<Wire.h>
#define I2C_ADDRESS 0x48
ADS1115_WE adc(I2C_ADDRESS);

#define WindSensorPin 2

#define MIN_ANALOG_VALUE  0     /* Minimum value of the ADC  => 0° */
#define MAX_ANALOG_VALUE  26352  /* Maximum value of the ADC  => 359.9°*/

long MIN_MAX_ANALOG_check_count = 0;
long MIN_ANALOG_VALUE_measure = MAX_ANALOG_VALUE;  
long MAX_ANALOG_VALUE_measure = MIN_ANALOG_VALUE;  


const int led = LED_BUILTIN;  // the pin with a LED

//**************** Infos about Windspeed **********************
// ***  max. delta between 2 pulses
// ***  derive then the debounce time
// 1600 pulse/hour = 1 mph  
// Max Count (@200mph) =  320,000 pulse/hour 
// 320000/(60*60) == 88.88888 (Hz) pulse/sec 
// Intervall 2.5 sec => 222.222 pusle/sec
// 88.88888 (Hz) pulse/sec 
// 1/88.888 => 11.25 msec for each pulse
// ***  debounce time < 11 msec!
#define debounceTime 10

// ************** Check Davis given formular ******************
// Davis: 1600 pulse/hour = 1 mph  
// 1600 / 3600 = 0.444444 pulse/sec = 1 mph
// 0,4444 * 2.5 = 1.1111 pulse/2.5sec = 1 mph
// Davis: v = p(2.25 / t)
// v = Speed in mph
// p = Pulses Count
// t = Time in sec
// 1 mph = 1600 (2.25/3600sec) => ok!

#define MEASURMENT_INTERVALL double(5)  /* Measruement period X sec, afetr X sec the result will be reported via the serial interface*/

// ************** Calculate the conversion factors for the units: mph, km/h and meter/sec******************
#define conversion_factor_ms   0.44704  /* mph *0.44704 => m/s */
#define conversion_factor_mph  1.0      /* mph * 1.0    => mph */  
#define conversion_factor_kmh  1.60934  /*mph * 1,60934 => km/h*/

#define SELECTED_CONVERSION conversion_factor_ms  

const double SpeedPerPulse = ((2.25 / MEASURMENT_INTERVALL) * SELECTED_CONVERSION ); // Pusles per Time * conversion_factor

void setup() 
{
  Wire.begin();
  Serial.begin(115200);

  /* Set the voltage range of the ADC to adjust the gain
   * Please note that you must not apply more than VDD + 0.3V to the input pins!
   * 
   * ADS1115_RANGE_6144  ->  +/- 6144 mV
   * ADS1115_RANGE_4096  ->  +/- 4096 mV
   * ADS1115_RANGE_2048  ->  +/- 2048 mV (default)
   * ADS1115_RANGE_1024  ->  +/- 1024 mV
   * ADS1115_RANGE_0512  ->  +/- 512 mV
   * ADS1115_RANGE_0256  ->  +/- 256 mV
   */
  adc.setVoltageRange_mV(ADS1115_RANGE_4096);
  adc.setCompareChannels(ADS1115_COMP_0_GND); 
  adc.setMeasureMode(ADS1115_CONTINUOUS);
  
  pinMode(WindSensorPin, INPUT);
  attachInterrupt(digitalPinToInterrupt(WindSensorPin), rotationCallback, FALLING); 

  pinMode(led, OUTPUT);    

  Timer1.initialize(MEASURMENT_INTERVALL*1000000UL); // Timer for each MEASURMENT_INTERVALL sec
  Timer1.attachInterrupt( MeasruementIntervallCB); // attach the service routine 
  
}

volatile unsigned long tvalue = millis();
volatile bool sendValuesSerial = false;

double windDirection_X_sum = 0.0;
double windDirection_Y_sum = 0.0;
int    windDirection_measurementCount = 0;


// ******* Infos about wind gust ********
//v = p(2.25 / t)
// v = 1 Pulse * (2.25 / delta_t (millisec!))    
// max 200mph = 1 * (2.25 /delta_t))
// gust@200mph => min delta_t = 11.25 ms  
// min 1mph = 1 * (2.25 /delta_t))
// gust@1mph   => max delta_t = 2250 ms    
#define maxDeltaTimeMeasurement 2250UL

volatile unsigned long windDirMeasurmentTime = 0; // keeps the last measurement time (Millis)
volatile unsigned long rotationCountRaw = 0;      // Keeps the number of pulses
volatile double calculatedWindSpeed     = 0.0;    
volatile double calculatedWindGust      = 0.0;

 
void loop()
{
  // Cacl. done in Timer task - ready to send wind speed and all other values
  if (sendValuesSerial)
  {
    sendValuesSerial = false;
    Serial.print("VALUE WGUST:");Serial.print(calculatedWindGust,  1);
    Serial.print(" m/s WSPEED:");Serial.print(calculatedWindSpeed, 1);
    Serial.print(" m/s WDIR:");

    double windDirection_X_avg = 0.0;
    double windDirection_Y_avg = 0.0;

    if (windDirection_measurementCount > 0)
    {
      windDirection_X_avg = windDirection_X_sum / windDirection_measurementCount;
      windDirection_Y_avg = windDirection_Y_sum / windDirection_measurementCount;
    }
    windDirection_measurementCount = 0;
    windDirection_X_sum =0.0;
    windDirection_Y_sum =0.0;

    
    if ((windDirection_X_avg == 0.0) && (windDirection_Y_avg  == 0.0)) // No Wind -> no Wind direction
    {
      Serial.print("UNDEF");
    }
    else
    {
      double calculatedWindDirection = atan2(abs(windDirection_X_avg), abs(windDirection_Y_avg))* 180.0 / M_PI; // atan2 considers the illegal arctan(0) call
      if ((windDirection_X_avg >= 0) && (windDirection_Y_avg >= 0)) calculatedWindDirection = calculatedWindDirection;
      if ((windDirection_X_avg >= 0) && (windDirection_Y_avg < 0 )) calculatedWindDirection = 180 - calculatedWindDirection;
      if ((windDirection_X_avg < 0 ) && (windDirection_Y_avg >= 0)) calculatedWindDirection = 360 - calculatedWindDirection;
      if ((windDirection_X_avg < 0 ) && (windDirection_Y_avg < 0 )) calculatedWindDirection = 180 + calculatedWindDirection;
      Serial.print(calculatedWindDirection, 1);
    }  
    Serial.print(" DEG\n");      

    MIN_MAX_ANALOG_check_count ++;
    if (MIN_MAX_ANALOG_check_count > 2000)  // Check every MEASURMENT_INTERVALL * 2000 the min/max raw values of the ADC  
    {
      if ((abs(MAX_ANALOG_VALUE_measure - MAX_ANALOG_VALUE) > 20) || (abs(MIN_ANALOG_VALUE_measure - MIN_ANALOG_VALUE) > 20))    // If there are major deviation inform the host system
      {
         Serial.print("ALIGN:");Serial.print(MIN_ANALOG_VALUE_measure);Serial.print("min;");Serial.print(MAX_ANALOG_VALUE_measure);Serial.print("MAX\n");
        
      }      
      MIN_MAX_ANALOG_check_count = 0;
      MIN_ANALOG_VALUE_measure = MAX_ANALOG_VALUE;  
      MAX_ANALOG_VALUE_measure = MIN_ANALOG_VALUE;  
    }
  
  }

  unsigned long now = millis();
  if((now - windDirMeasurmentTime ) > 500)  // Measure Wind Direction every half second
  {
    windDirMeasurmentTime = now;
    
    long rawValue = adc.getRawResult();
    
    if (rawValue  < MIN_ANALOG_VALUE_measure)
       MIN_ANALOG_VALUE_measure =  rawValue;
    
    if (rawValue > MAX_ANALOG_VALUE_measure)
      MAX_ANALOG_VALUE_measure =  rawValue;
  

    double dir = mapf(constrain(rawValue, MIN_ANALOG_VALUE, MAX_ANALOG_VALUE), MIN_ANALOG_VALUE, MAX_ANALOG_VALUE, 0.0, 359.9) / 180.0 * M_PI; // in radian 180.0 * M_PI

    // calculatedWindSpeed from the last calc 
    windDirection_X_sum = windDirection_X_sum + (calculatedWindSpeed*sin(dir));
    windDirection_Y_sum = windDirection_Y_sum + (calculatedWindSpeed*cos(dir));
    windDirection_measurementCount ++;  
  }
}

double mapf(double x, double in_min, double in_max, double out_min, double out_max)
{
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}


volatile unsigned long minDeltaTimeMeasurement = maxDeltaTimeMeasurement ;

volatile unsigned long debounceRotation = 0;
volatile int ledState = 1;

volatile unsigned long deltaTimeMeasurement = 0; // calc Time between 2 pules (in order to get values to calc wind gust)


void rotationCallback() // ISR
{
  unsigned long now = millis();
  if((now - debounceRotation ) > debounceTime) 
  {
    rotationCountRaw ++;
    debounceRotation = now;
    
    if ((now - deltaTimeMeasurement) < minDeltaTimeMeasurement ) // Time between 2 pules smaller than old value, store new minimum
       minDeltaTimeMeasurement = now - deltaTimeMeasurement;
      
    deltaTimeMeasurement = now;
    
    digitalWrite(led, ledState);
    ledState = !ledState;
  }
}


void MeasruementIntervallCB()
{
  sendValuesSerial = true;
  calculatedWindSpeed = double(rotationCountRaw) * SpeedPerPulse;
  
  if (minDeltaTimeMeasurement < maxDeltaTimeMeasurement )
  {
    calculatedWindGust = ((2.25 / (double(minDeltaTimeMeasurement)/1000.0)))* SELECTED_CONVERSION;    
  }
  else
  {
    calculatedWindGust = 0.0;
  }    

  minDeltaTimeMeasurement = maxDeltaTimeMeasurement ;
  rotationCountRaw = 0;
}
